require("veafMissionEditor")

local DEFAULT_POSITION_X = 0
local DEFAULT_POSITION_Y = 0

---checks if a string starts with a prefix
---@param aString any
---@param aPrefix any
---@param caseSensitive? boolean   ; if true, case sensitive search
---@return boolean
function startsWith(aString, aPrefix, caseSensitive)
    local aString = aString
    if not aString then
        return false
    elseif not caseSensitive then
        aString = aString:upper()
    end
    local aPrefix = aPrefix
    if not aPrefix then
        return false
    elseif not caseSensitive then
        aPrefix = aPrefix:upper()
    end
    return string.sub(aString,1,string.len(aPrefix))==aPrefix
end

function deepcopy(obj, seen)
    -- Handle non-tables and previously-seen tables.
    if type(obj) ~= 'table' then return obj end
    if seen and seen[obj] then return seen[obj] end
  
    -- New table; mark it as seen and copy recursively.
    local s = seen or {}
    local res = {}
    s[obj] = res
    for k, v in pairs(obj) do res[deepcopy(k, s)] = deepcopy(v, s) end
    return setmetatable(res, getmetatable(obj))
end

local function processGroup(group)
  local rGroup = nil
  if group ~= nil then
    rGroup = deepcopy(group)
    if rGroup.name and not startsWith(rGroup.name, "veafSpawn-", false) then
      rGroup.name = "veafSpawn-" .. rGroup.name
    end
    rGroup.hidden = true
    rGroup.lateActivation = true
    rGroup.x = DEFAULT_POSITION_X
    rGroup.y = DEFAULT_POSITION_Y
    local route = rGroup.route
    if route then
      local points = route.points
      if points then
        for index, point in pairs(points) do
          if index ~= 1 then
            points[index] = nil -- remove all navigation points after the first one
          else
            points[index].x = DEFAULT_POSITION_X
            points[index].y = DEFAULT_POSITION_Y
          end
        end
      end
    end
    local units = rGroup.units
    if units then
      for _, unit in pairs(units) do
        if unit.name and not startsWith(unit.name, "veafSpawn-", false) then
          unit.name = "veafSpawn-" .. unit.name
        end
        unit.x = DEFAULT_POSITION_X
        unit.y = DEFAULT_POSITION_Y
      end
    end
  end
  return rGroup
end

local function processFunction(missionTable)
  local rMissionTable = {}
  for settingGroupName, settingGroup in pairs(missionTable) do
    local rSettingGroup = {}
    rSettingGroup["coalition"] = settingGroup["coalition"]
    rSettingGroup["country"] = settingGroup["country"]
    rSettingGroup["category"] = settingGroup["category"]
    rSettingGroup["groups"] = {}
    local rSettingGroups = rSettingGroup["groups"]
    local settingGroups = settingGroup["groups"]
    if settingGroups then
      for _, group in pairs(settingGroups) do       
        table.insert(rSettingGroups, processGroup(group))
      end
    end
    rMissionTable[settingGroupName] = rSettingGroup
  end
  return rMissionTable
end

veafMissionEditor.editMission(
  [[D:\dev\_VEAF\VEAF-Mission-Creation-Tools\src\defaults\veafSpawnableAircraftsEditor\settings.lua]], 
  [[D:\dev\_VEAF\VEAF-Mission-Creation-Tools\src\defaults\veafSpawnableAircraftsEditor\settings.processed.lua]], 
  "settings", processFunction, false)